#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "aspnger.h"

int debug = 0;
int flags = 0;
struct image *target;
struct list_head hash_list[256];
char *output_filename = NULL;

int get_opts(int *argc, char **argv[])
{
  while (*argc && (*argv)[0][0] == '-') {
    if (strcmp((*argv)[0], "-k8") == 0)
      flags |= KARTTAPAIKKA8000;
    else if (strcmp((*argv)[0], "-k40") == 0)
      flags |= KARTTAPAIKKA40000;
    else if (strcmp((*argv)[0], "-nc") == 0)
      flags |= NOT_COLORIZE_BROWN;
    else
      return -1;
    (*argc)--;
    (*argv)++;
  }
  return 0;
}

int main(int argc, char *argv[])
{
  char *output_filename;
  int i,j;

  argv++; argc--;

  for(i=0;i<256;i++)
    INIT_LIST_HEAD(&hash_list[i]);

  if (get_opts(&argc, &argv) || argc < 2) {
    printf("Use: aspnger [-k8|-k40] output.png file.png...\n");
    return 1;
  }

  output_filename = argv[0];
  argv++; argc--;

  /* Initial file */
  if (!(target = LoadPNG(argv[0])))
      return 1;
  add_hashes(target, hash_list);
  argv++; argc--;

  /* main stitching loop */
  while (argc) {
    for(i = 0, j = 0; i < argc; i++)
    {
      struct image *source;

      if (!(source = LoadPNG(argv[i])))
        return 1;

      if (debug) fprintf(stderr, "%s: ", argv[i]); fflush(stderr);

      if (MATCH(source, target, hash_list))
      {
        MERGE(target, source);
        add_hashes(source, hash_list);
        DeletePNGrows(source);
        if (debug) fprintf(stderr, "ok.\n");
      }
      else
      {
        if (debug) fprintf(stderr, "pending...\n");
        argv[j++] = argv[i];
      }
    }

    if (j == i) {
      fprintf(stderr, "ERROR: Not matched files:\n");
      for (i=0; i<j; i++) printf("  %s\n", argv[i]);
      exit(1);
    }

    argc = j;
  }

  /* Filtering output */
  if (flags & KARTTAPAIKKA8000) {
    clean_karttapaikka8000(target);
    target->res = 10000; /* per meter */
  }
  if (flags & KARTTAPAIKKA40000) {
    init_k40000();
    clean_karttapaikka40000(target);
    target->res = 5000; /* per meter */
  }
  if (flags && !(flags & NOT_COLORIZE_BROWN))
    colorize_brown(target);

  /* Save */
  return SavePNG(target, output_filename);
}
