
// #define deb(fmt, args...) printf("%s: " fmt, __FUNCTION__, ##args)
#define deb(fmt, args...) while(0) {}

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>
#include "aspnger.h"


u32 fletcher(struct image *img, long x, long y)
{
  long  i, j, xx;
  u32   sum1 = 0xffff, sum2 = 0xffff; /* Fletcher's checksum */
  int   count = 0;

  for (i = 0; i < HASH_SZ; i++, y++) {
    for (j = 0, xx=x*3; j < HASH_SZ; j++, xx+=3) {
      sum1 += img->data[y][xx+0]; sum2 += sum1;
      sum1 += img->data[y][xx+1]; sum2 += sum1;
      sum1 += img->data[y][xx+2]; sum2 += sum1;
      count += 3;
      if (count >= 360) {
        sum1 = (sum1 & 0xffff) + (sum1 >> 16);
        sum2 = (sum2 & 0xffff) + (sum2 >> 16);
        count -= 360;
      }
    }
  }

  sum1 = (sum1 & 0xffff) + (sum1 >> 16);
  sum2 = (sum2 & 0xffff) + (sum2 >> 16);
  return sum2 << 16 | sum1;
}

#define max(a,b) ((a)>=(b) ? (a) : (b))
#define min(a,b) ((a)<=(b) ? (a) : (b))

int verify_match(struct image *src, struct image *other, struct image *target)
{
  long x1, x2, y1, y2; /* rect in orig coord */
  long x, y; /* orig coord */
  long sx, sy; /* source coord */
  long tx, ty; /* target coord */

  deb("other xoff=%ld yoff=%ld, width=%ld height=%ld\n",
      other->xoff, other->yoff, other->width, other->height);

  x1 = max(src->xoff, other->xoff);
  x2 = min(src->xoff + src->width, other->xoff + other->width);
  y1 = max(src->yoff, other->yoff);
  y2 = min(src->yoff + src->height, other->yoff + other->height);

  deb("x1..x2=%ld-%ld y1..y2=%ld-%ld\n", x1, x2, y1, y2);

  if (!(x2 > x1) || !(y2 > y1))
    return 0;

  for (y=y1, sy=(y1 - src->yoff), ty=(y1 - target->yoff);
       y < y2; y++, sy++, ty++) {
    for (x=x1, sx=3*(x1 - src->xoff), tx=3*(x1 - target->xoff);
         x < x2; x++, sx+=3, tx+=3) {
      if (src->data[sy][sx+0] != target->data[ty][tx+0] ||
          src->data[sy][sx+1] != target->data[ty][tx+1] ||
          src->data[sy][sx+2] != target->data[ty][tx+2])
      {
        deb("diff @ xoff=%ld yoff=%ld (src x=%ld y=%ld) (other x=%ld y=%ld)\n",
            x, y, sx/3, sy, tx/3, ty);
        return 0;
      }
    }
  }
  return 1;
}

int match(struct image *src, long x, long y, struct image *target, struct list_head *hash_list)
{
  u32 srchash;
  u8 hash_index;
  struct hash* candidate;

  srchash = fletcher(src, x, y);
  hash_index = (u8)srchash;

  // deb("hash=%08lx @ x=%ld y=%ld\n", srchash, x, y);
  list_for_each_entry(candidate, &hash_list[hash_index], list)
  {
    if (candidate->hash == srchash) {
      deb("CANDIDATE x=%ld y=%ld (@ xoff=%ld yoff=%ld)\n",
          candidate->x, candidate->y,
          candidate->img_info->xoff, 
          candidate->img_info->yoff);
      src->xoff = candidate->img_info->xoff + candidate->x - x;
      src->yoff = candidate->img_info->yoff + candidate->y - y;
      deb("src => xoff=%ld yoff=%ld, width=%ld height=%ld\n",
          src->xoff, src->yoff, src->width, src->height);

      if (verify_match(src, candidate->img_info, target))
        return 1;
    }
  }

  return 0;
}

int MATCH(struct image *src, struct image *target, struct list_head *hash_list)
{
  long x, y;

  deb("--- try top-left corner ---%s\n", "");
  /* try top-left corner */
  for (x = 0; x < HASH_SZ; x++)
    for (y = 0; y < HASH_SZ; y++)
      if (match(src, x, y, target, hash_list)) {
        deb("MATCH TOP LEFT x=%ld y=%ld\n", x, y);
        return 1;
      }

  deb("--- try top-right corner ---%s\n", "");
  /* try top-right corner */
  for (x = 0; x < HASH_SZ; x++)
    for (y = 0; y < HASH_SZ; y++)
      if (match(src, (src->width - HASH_SZ - x), y, target, hash_list)) {
        deb("MATCH TOP RIGHT x=%ld y=%ld\n", x, y);
        return 1;
      }

  deb("--- try bottom-left corner ---%s\n", "");
  /* try bottom-left corner */
  for (x = 0; x < HASH_SZ; x++)
    for (y = 0; y < HASH_SZ; y++)
      if (match(src, x, (src->height - HASH_SZ - y), target, hash_list)) {
        deb("MATCH BOTTOM LEFT x=%ld y=%ld\n", x, y);
        return 1;
      }

  deb("--- try bottom-right corner ---%s\n", "");
  /* try bottom-right corner */
  for (x = 0; x < HASH_SZ; x++)
    for (y = 0; y < HASH_SZ; y++)
      if (match(src, (src->width - HASH_SZ - x), (src->height - HASH_SZ - y), target, hash_list)) {
        deb("MATCH BOTTOM RIGHT x=%ld y=%ld\n", x, y);
        return 1;
      }

  deb("%s\n", "nope");

  return 0;
}

void enlarge(struct image *target, struct image *src) {
  long offset;
  int i;

  /* make space to left */
  offset = target->xoff - src->xoff;
  if (offset > 0) {
    deb("to left %ld\n", offset);
    for(i = 0; i < target->height; i++) {
      target->data[i] = realloc(target->data[i], (target->width + offset)*3);
      memmove(&target->data[i][offset*3], &target->data[i][0], target->width*3);
      memset(&target->data[i][0], 0xff, offset*3);
    }
    target->xoff = src->xoff;
    target->width += offset;
  }

  /* make space to right */
  offset = (src->width + src->xoff) - (target->width + target->xoff);
  if (offset > 0) {
    deb("to right %ld\n", offset);
    for(i = 0; i < target->height; i++) {
      target->data[i] = realloc(target->data[i], (target->width + offset)*3);
      memset(&target->data[i][target->width*3], 0xff, offset*3);
    }
    target->width += offset;
  }

  /* make space to top */
  offset = target->yoff - src->yoff;
  if (offset > 0) {
    deb("to top %ld\n", offset);
    target->data = realloc(target->data, (target->height + offset)*sizeof(u8 *));
    for(i = target->height-1; i >= 0; i--) /* move */
      target->data[i+offset] = target->data[i];
    for(i = 0; i < offset; i++) {
      target->data[i] = malloc(target->width * 3);
      memset(target->data[i], 0xff, target->width * 3);
    }
    target->yoff = src->yoff;
    target->height += offset;
  }

  /* make space to bottom */
  offset = (src->height + src->yoff) - (target->height + target->yoff);
  if (offset > 0) {
    deb("to bottom %ld\n", offset);
    target->data = realloc(target->data, (target->height + offset)*sizeof(u8 *));
    for(i = 0; i < offset; i++) {
      target->data[target->height+i] = malloc(target->width * 3);
      memset(target->data[target->height+i], 0xff, target->width * 3);
    }
    target->height += offset;
  }
}

/* origs must match */
void MERGE(struct image *target, struct image *src) {
  long x, y; /* source */
  long tx, ty; /* target */

  enlarge(target, src);

  for(y = 0, ty = (src->yoff - target->yoff);
      y < src->height; y++, ty++) {
    for(x = 0, tx = 3*(src->xoff - target->xoff);
        x < 3*src->width; x+=3, tx+=3) {
      target->data[ty][tx+0] = src->data[y][x+0];
      target->data[ty][tx+1] = src->data[y][x+1];
      target->data[ty][tx+2] = src->data[y][x+2];
    }
  }
}

void add_hash(struct image *img, long x, long y, struct list_head *hash_list)
{
  u8 hash_index;
  struct hash *hash;

  hash = malloc(sizeof(struct hash));
  hash->img_info = img;
  hash->x = x;
  hash->y = y;
  hash->hash = fletcher(img, x, y);

  hash_index = (u8)hash->hash;
//  deb("x=%ld y=%ld (@ xoff=%ld yoff=%ld) hash=%08lx hash_index=0x%02x\n", 
//      x, y, img->xoff, img->yoff, hash->hash, hash_index);
  list_add(&hash->list, &hash_list[hash_index]);
}

void add_hashes(struct image *img, struct list_head *hash_list)
{
  long x,y;
  long xmid = img->width/2;
  long ymid = img->height/2;

  /* top-left quarter */
  for(y = 0; y < ymid; y += HASH_SZ)
    for(x = 0; x < xmid; x += HASH_SZ)
      add_hash(img, x, y, hash_list);

  /* top-right quarter */
  for(y = 0; y < ymid; y += HASH_SZ)
    for(x = (img->width - HASH_SZ); x > (xmid + 1 - HASH_SZ); x -= HASH_SZ)
      add_hash(img, x, y, hash_list);

  /* bottom-left quarter */
  for(y = (img->height - HASH_SZ); y > (ymid + 1 - HASH_SZ); y -= HASH_SZ)
    for(x = 0; x < xmid; x += HASH_SZ)
      add_hash(img, x, y, hash_list);

  /* bottom-right quarter */
  for(y = (img->height - HASH_SZ); y > (ymid + 1 - HASH_SZ); y -= HASH_SZ)
    for(x = (img->width - HASH_SZ); x > (xmid + 1 - HASH_SZ); x -= HASH_SZ)
      add_hash(img, x, y, hash_list);

}

